<?php

namespace Drupal\commerce_shipping_best_rate\Plugin\Commerce\ShippingMethod;

use Drupal\commerce_price\Price;
use Drupal\commerce_shipping\Entity\ShipmentInterface;
use Drupal\commerce_shipping\PackageTypeManagerInterface;
use Drupal\commerce_shipping\Plugin\Commerce\ShippingMethod\ShippingMethodBase;
use Drupal\commerce_shipping\ShippingRate;
use Drupal\commerce_shipping\ShippingService;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\state_machine\WorkflowManagerInterface;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides the BestRate shipping method.
 *
 * @CommerceShippingMethod(
 *   id = "best_rate",
 *   label = @Translation("Best rate"),
 * )
 */
class BestRate extends ShippingMethodBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new BestRate object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\commerce_shipping\PackageTypeManagerInterface $package_type_manager
   *   The package type manager.
   * @param \Drupal\state_machine\WorkflowManagerInterface $workflow_manager
   *   The workflow manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, PackageTypeManagerInterface $package_type_manager, WorkflowManagerInterface $workflow_manager, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $package_type_manager, $workflow_manager);

    $this->entityTypeManager = $entity_type_manager;
    $this->services['default'] = new ShippingService('default', $this->configuration['rate_label']);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.commerce_package_type'),
      $container->get('plugin.manager.workflow'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Get the array of service ids to group together.
   *
   * @return array
   *   The service ids.
   */
  public function getServicesGroup() {
    return $this->configuration['services_group'];
  }

  /**
   * Get the array of roles that should be excluded from best rate.
   *
   * @return array
   *   The roles.
   */
  public function getExcludedRoles() {
    return $this->configuration['excluded_roles'];
  }

  /**
   * Get the display option for excluded roles.
   *
   * @return bool
   *   The display option.
   */
  public function getExcludedRolesDisplay() {
    return $this->configuration['excluded_roles_display'];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'rate_label' => '',
      'rate_description' => '',
      'services' => ['default'],
      'services_group' => [],
      'excluded_roles' => [],
      'excluded_roles_display' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Collect all of the configured shipping methods.
    // @todo Filter by store.
    $shipping_method_storage = $this->entityTypeManager->getStorage('commerce_shipping_method');
    $shipping_methods_query = $shipping_method_storage->getQuery()
      ->accessCheck(TRUE)
      ->condition('status', 1)
      ->condition('plugin', 'best_rate', '<>')
      ->sort('shipping_method_id', 'ASC')
      ->execute();
    /** @var \Drupal\commerce_shipping\Entity\ShippingMethodInterface[] $shipping_methods */
    $shipping_methods = $shipping_method_storage->loadMultiple($shipping_methods_query);

    // Create a list of enabled services.
    $services = [];
    foreach ($shipping_methods as $shipping_method) {
      $shipping_method_id = $shipping_method->id();
      $shipping_method_label = $shipping_method->getName();
      foreach ($shipping_method->getPlugin()->getServices() as $service) {
        // The value follows the format "shipping_method_id--service_id" which
        // will match the rate id when filtering rates in the event subscriber.
        $services[$shipping_method_id . '--' . $service->getId()] = $shipping_method_label . ' - ' . $service->getLabel();
      }
    }

    // Load all available roles.
    $roles = array_map(function (RoleInterface $role) {
      return $role->label();
    }, Role::loadMultiple());

    $form['rate_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rate label'),
      '#description' => $this->t('Shown to customers when selecting the rate.'),
      '#default_value' => $this->configuration['rate_label'],
      '#required' => TRUE,
    ];
    $form['rate_description'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rate description'),
      '#description' => $this->t('Provides additional details about the rate to the customer.'),
      '#default_value' => $this->configuration['rate_description'],
    ];
    $form['services_group'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Services to group'),
      '#options' => $services,
      '#default_value' => $this->configuration['services_group'],
      '#required' => TRUE,
    ];
    $form['exclude_role'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Roles to exclude'),
      '#description' => $this->t('Roles which should see regular rates instead of best rates.'),
    ];
    $form['exclude_role']['excluded_roles'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Excluded roles'),
      '#options' => array_map('\Drupal\Component\Utility\Html::escape', $roles),
      '#default_value' => $this->configuration['excluded_roles'],
    ];
    $form['exclude_role']['excluded_roles_display'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show both regular rates and this best rate for excluded roles.'),
      '#default_value' => $this->configuration['excluded_roles_display'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['rate_label'] = $values['rate_label'];
      $this->configuration['rate_description'] = $values['rate_description'];
      if (!empty($values['services_group'])) {
        $values['services_group'] = array_filter($values['services_group']);

        $this->configuration['services_group'] = array_keys($values['services_group']);
      }
      if (!empty($values['exclude_role']['excluded_roles'])) {
        $values['exclude_role']['excluded_roles'] = array_filter($values['exclude_role']['excluded_roles']);

        $this->configuration['excluded_roles'] = array_keys($values['exclude_role']['excluded_roles']);
      }
      $this->configuration['excluded_roles_display'] = $values['exclude_role']['excluded_roles_display'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function calculateRates(ShipmentInterface $shipment) {
    $rates = [];
    $rates[] = new ShippingRate([
      'shipping_method_id' => $this->parentEntity->id(),
      'service' => $this->services['default'],
      'amount' => new Price('0.00', 'USD'),
      'description' => $this->configuration['rate_description'],
    ]);

    return $rates;
  }

}
